//*****************************************************************************
//*****************************************************************************
//  FILENAME:   sensor.c
//  Version: 1.0, 2007/03/08
//  Generated by Andy Kiser
//
//  DESCRIPTION:  sensor specific functions 
//
//-----------------------------------------------------------------------------
// Originally authored by Toradex AG. This work is now available under the terms and conditions
// of the Creative Commons License 'Attribution CC BY', details of which can be found here:
// http://creativecommons.org/licenses/by/3.0/
//*****************************************************************************
//*****************************************************************************
//#define __DEBUG__
#include <m8c.h>        // part specific constants and macros
#include "sensor.h"
#include "psocgpioint.h"
#include "feature.h"
#include "parameters.h"
#include "types_n_macros.h"
#include "PSoCAPI.h"    // PSoC API definitions for all User Modules
#include "PSoCDynamic.h"	//Dynamic Load of Configurations

//----------------------------------------
// Sensor parameters
// Factory Defaults are defined here (NUM_xxx are defined in "sensor.h")
//----------------------------------------
const BYTE FactoryDefault_Param_8   [NUM_PARAM_8] =
{
	0x00,		// report mode
	0x09,		// LED mode
	0x02,		// Standart Sensitivity	//!02
	0x00,		// Shield mode Set P0[4] as sensor
	0x28,		// Finger Threshold
	0x0c,		// Resoulution: 12 bit
	0x01,		// Scan Speed Fast
	0x01		// Auto Base Line ON
};

const WORD FactoryDefault_Param_16  [NUM_PARAM_16] = 
{
	50,				// report rate [ms]
	50,				// sample rate [ms]
	50000,			// sample rate [us]	(not used for this sensor)
	0xFFFF,			// Baseline 0
	0xFFFF,			// Baseline 1
	0xFFFF,			// Baseline 2
	0xFFFF,			// Baseline 3
};

const DWORD FactoryDefault_Param_32  [NUM_PARAM_32] =
{
	0x00000000	// currently no 32bit parameters
};

const tParam_168 FactoryDefault_Param_168 [NUM_PARAM_168] =
{
	"Isolated Input",	// user device name
	"Frame Number",		// user name Channel 0
	"Buttons",			// user name Channel 1
	"Proximity Ch 0",	// user name Channel 2
	"Proximity Ch 1",	// user name Channel 3
	"Proximity Ch 2",	// user name Channel 4
	"Proximity Ch 3",	// user name Channel 5
};


//----------------------------------------
// this module's global variables
//----------------------------------------
tReport Report;
extern BYTE SOF_Flag;	// Set in SOF_ISR

//----------------------------------------
// constant lookup-table for ADC->physical value conversion
//----------------------------------------


//----------------------------------------
// type definitions
//----------------------------------------

typedef enum 
{	
	eIdle
} tFSM_Measure;

typedef enum 
{	
	eOff,
	eGoOn,
	eOn,
	eGoOff
} tFSM_Keys;

//----------------------------------------
// variables

tFSM_Measure	FSM;
WORD			Countdown_1ms;

BOOL			NewSample = FALSE;
tFSM_Keys		FSM_Keys[CSD_1_TotalSensorCount];
BYTE			DebounceCount[CSD_1_TotalSensorCount];

// store original drivemodes during suspend
BYTE			DM0_2, DM0_1, DM0_0;
BYTE			DM1_2, DM1_1, DM1_0;
BYTE			DM2_2, DM2_1, DM2_0;
BYTE			DM3_2, DM3_1, DM3_0;
BYTE			DM5_2, DM5_1, DM5_0;

void DetectAllButtons(void);

//-----------------------------------------------------------------------------
// Set the sensitivity of the sensor by changing the Modulation Capacitor 
// and the Feedback Resistor
//-----------------------------------------------------------------------------
void Set_Sensitivity()
{
	BYTE Sensitivity;
	//BYTE Test;
	Get_Param_8(PAR8_SENSITIVITY, VOLATILE, &Sensitivity);
	if (Sensitivity == 0x00)	
	{
		//Modulation Capacitor Pin P0[5]
		MUX_CR0 &= 0x7F;	// Clear bit 7 (Pin0[7])
		MUX_CR0 |= 0x20;	// Set bit for Pin0[5]
		AMUXCFG |= 0x10;	//Sets bit 5 and 4 to 0b01
		AMUXCFG &= 0xDF;

		//Feedback Resistor Pin P3[5]
		PRT1DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT3DM0 &= 0xDD;	//Sets other feedback pins to analog in
		PRT5DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT3DM0 |= 0x20;	//Sets Pin P35 to Open Drain Low
		PRT1GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT3GS	&= 0xDD;	//Clears other GlobalOutOdd settings
		PRT5GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT3GS	|= 0x20;	//Sets Pin P35 to GlobalOutOdd
		CMP_GO_EN &= 0x3F;	//Clears old compare bus settings
		CMP_GO_EN |= 0x80;	//Set compare to GOO5
	}
	else if (Sensitivity == 0x01)	
	{
		//Modulation Capacitor Pin P0[5]
		MUX_CR0 &= 0x7F;	// Clear bit 7 (Pin0[7])
		MUX_CR0 |= 0x20;	// Set bit for Pin0[5]
		AMUXCFG |= 0x10;	//Sets bit 5 and 4 to 0b01
		AMUXCFG &= 0xDF;

		//Feedback Resistor Pin P3[1]
		PRT1DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT3DM0 &= 0xDD;	//Sets other feedback pins to analog in
		PRT5DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT3DM0 |= 0x02;	//Sets Pin P31 to Open Drain Low
		PRT1GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT3GS	&= 0xDD;	//Clears other GlobalOutOdd settings
		PRT5GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT3GS	|= 0x02;	//Sets Pin P31 to GlobalOutOdd
		CMP_GO_EN &= 0x3F;	//Clears old compare bus settings
		CMP_GO_EN |= 0x40;	//Set compare to GOO1
	}
	else if (Sensitivity == 0x02)	
	{
		//Modulation Capacitor Pin P0[7]
		MUX_CR0 &= 0xDF;	// Clear bit 5 (Pin0[5])
		MUX_CR0 |= 0x80;	// Set bit for Pin0[7]
		AMUXCFG |= 0x20;	//Sets bit 5 and 4 to 0b10
		AMUXCFG &= 0xEF;

		//Feedback Resistor Pin P5[5]
		PRT1DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT3DM0 &= 0xDD;	//Sets other feedback pins to analog in
		PRT5DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT5DM0 |= 0x20;	//Sets Pin P55 to Open Drain Low
		PRT1GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT3GS	&= 0xDD;	//Clears other GlobalOutOdd settings
		PRT5GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT5GS	|= 0x20;	//Sets Pin P55 to GlobalOutOdd
		CMP_GO_EN &= 0x3F;	//Clears old compare bus settings
		CMP_GO_EN |= 0x80;	//Set compare to GOO5
	}
	else if (Sensitivity == 0x03)	
	{
		//Modulation Capacitor Pin P0[7]
		MUX_CR0 &= 0xDF;	// Clear bit 5 (Pin0[5])
		MUX_CR0 |= 0x80;	// Set bit for Pin0[7]
		AMUXCFG |= 0x20;	//Sets bit 5 and 4 to 0b10
		AMUXCFG &= 0xEF;

		//Feedback Resistor Pin P1[5]
		PRT1DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT3DM0 &= 0xDD;	//Sets other feedback pins to analog in
		PRT5DM0 &= 0xDF;	//Sets other feedback pins to analog in
		PRT1DM0 |= 0x20;	//Sets Pin P15 to Open Drain Low
		PRT1GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT3GS	&= 0xDD;	//Clears other GlobalOutOdd settings
		PRT5GS	&= 0xDF;	//Clears other GlobalOutOdd settings
		PRT1GS	|= 0x20;	//Sets Pin P15 to GlobalOutOdd
		CMP_GO_EN &= 0x3F;	//Clears old compare bus settings
		CMP_GO_EN |= 0x80;	//Set compare to GOO5
	}
}

//-----------------------------------------------------------------------------
// Sets the Pin P0[4] as sensor input or as shield output
//-----------------------------------------------------------------------------
void Set_Shield()
{
	BYTE Shield;
	//BYTE Test;
	Get_Param_8(PAR8_SHIELD, VOLATILE, &Shield);
	if (Shield == 0x00)	//	Sets pin as sensor input
	{
		PRT0DM0 &= 0xEF;
		PRT0DM1 |= 0x10;
		PRT0DM2 |= 0x10;
		PRT0GS  &= 0xEF;
		MUX_CR0 &= 0xEF;
	}
	else if (Shield == 0x01)	//Sets pin as shield output
	{
		PRT0DM0 |= 0x10;
		PRT0DM1 &= 0xEF;
		PRT0DM2 &= 0xEF;
		PRT0GS  |= 0x10;
		MUX_CR0 |= 0x10;
	}
}

//-----------------------------------------------------------------------------
// Sets the finger threshold for all channels
//-----------------------------------------------------------------------------
void Set_Finger_Threshold()
{
	BYTE Threshold,i;
	Get_Param_8(PAR8_FINGER_THRESHOLD , VOLATILE, &Threshold);
	for(i=0;i<CSD_1_TotalSensorCount;i++)
	{
		CSD_1_baBtnFThreshold[i] = Threshold;
	}
}

//-----------------------------------------------------------------------------
// Sets the Resolution
//-----------------------------------------------------------------------------
void Set_Resolution()
{
	BYTE Resolution,Speed;
	Get_Param_8(PAR8_RESOLUTION, VOLATILE, &Resolution);
	Get_Param_8(PAR8_SCAN_SPEED , VOLATILE, &Speed);

	CSD_1_SetScanMode(Speed,Resolution);
}

//-----------------------------------------------------------------------------
// Sets the Baseline
//-----------------------------------------------------------------------------
void Set_BaseLine()
{
	WORD Baseline;
	BYTE i;
	for(i=0;i<CSD_1_TotalSensorCount;i++)
	{
		Get_Param_16(PAR16_SET_BASELINE_0+i, VOLATILE, &Baseline);
		if (Baseline == 0xFFFF)
		{
			CSD_1_InitializeSensorBaseline(i);
		}
		else
		{
			CSD_1_waSnsBaseline[i] = Baseline;
		}
	}
}

//-----------------------------------------------------------------------------
// Initializes the Sensor 
//-----------------------------------------------------------------------------
void Sensor_Start ()
{
	FSM = eIdle;
	Countdown_1ms = 0;
	// turn on power for the optocouper inputs
	//GPIO_HIGH(ON_nOFF)
	//Set_Sensitivity();
	Set_Shield();
	CSD_1_Start();
	//CSD_1_SetDefaultFingerThresholds();  
	//CSD_1_InitializeBaselines();
	Set_Sensitivity();
	Set_BaseLine(); 
	Set_Finger_Threshold();
	Set_Resolution(); 
	CSD_1_baSnsOnMask[0]=0;	//Set all Keys to 0
}
//-----------------------------------------------------------------------------
// Prepare Sensor for Suspend
//-----------------------------------------------------------------------------
void Sensor_Suspend()
{
	// turn off power for the sensor
	//GPIO_LOW(ON_nOFF)
	//CSD_1_Stop();

	// store original drivemodes during  suspend
	DM0_2 = PRT0DM2;
	DM0_1 = PRT0DM1;
	DM0_0 = PRT0DM0;

	DM1_2 = PRT1DM2;
	DM1_1 = PRT1DM1;
	DM1_0 = PRT1DM0;

	DM2_2 = PRT2DM2;
	DM2_1 = PRT2DM1;
	DM2_0 = PRT2DM0;

	DM3_2 = PRT3DM2;
	DM3_1 = PRT3DM1;
	DM3_0 = PRT3DM0;

	DM5_2 = PRT5DM2;
	DM5_1 = PRT5DM1;
	DM5_0 = PRT5DM0;
	
	// set all I/Os to/from the sensor to High-Z Analog (=110), 
	// except the OFF_nON signal. 
	// OFF_nON is on Port0, all other signals on Port2
	PRT0DM2  = 0xFF;
	PRT0DM1  = 0xFF;
	PRT0DM0  = 0x00;

	PRT1DM2  = 0xFF;
	PRT1DM1  = 0xFF;
	PRT1DM0  = 0x00;

	PRT2DM2  = 0xFF;
	PRT2DM1  = 0xFF;
	PRT2DM0  = 0x00;

	PRT3DM2  = 0xFF;
	PRT3DM1  = 0xFF;
	PRT3DM0  = 0x00;

	PRT5DM2  = 0xFF;
	PRT5DM1  = 0xFF;
	PRT5DM0  = 0x00;
}
//-----------------------------------------------------------------------------
// Resume Sensor from Suspend mode
//-----------------------------------------------------------------------------
void Sensor_Resume()
{
	BYTE ReadMode;
	
	// restore drivemodes for GPIOs to/from the sensor
	PRT0DM2 = DM0_2;
	PRT0DM1 = DM0_1;
	PRT0DM0 = DM0_0;

	PRT1DM2 = DM1_2;
	PRT1DM1 = DM1_1;
	PRT1DM0 = DM1_0;

	PRT2DM2 = DM2_2;
	PRT2DM1 = DM2_1;
	PRT2DM0 = DM2_0;

	PRT3DM2 = DM3_2;
	PRT3DM1 = DM3_1;
	PRT3DM0 = DM3_0;

	PRT5DM2 = DM5_2;
	PRT5DM1 = DM5_1;
	PRT5DM0 = DM5_0;

	// start up sensor again
	Sensor_Start();
}


//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// cannot be handled with the standard functionality in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//
//	return Value: number of Parameter bytes (1-31) that are 
//				  returned in the IN report
//-----------------------------------------------------------------------------
BYTE Sensor_Specific_Feature(void)
{
	// Initialize Baseline
	if  ( (!FEATURE_RPT.OUT.Get_nSet) && (FEATURE_RPT.OUT.Target == eAction) &&
	      (FEATURE_RPT.OUT.ParamSize == 0) && (FEATURE_RPT.OUT.Index_LSB == 0x01) )
		CSD_1_InitializeBaselines();
	return 1;
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// sets / gets a 8/16/32/168bit parameter in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//-----------------------------------------------------------------------------
void Sensor_Specific_Parameter(BYTE *NumberOfReturnedBytes)
{
	if  ( (!FEATURE_RPT.OUT.Get_nSet) && (FEATURE_RPT.OUT.ParamSize == 1) &&
	      (FEATURE_RPT.OUT.Index_LSB == PAR8_SENSITIVITY))
		Set_Sensitivity();
	else if  ( (!FEATURE_RPT.OUT.Get_nSet) && (FEATURE_RPT.OUT.ParamSize == 1) &&
	      (FEATURE_RPT.OUT.Index_LSB == PAR8_SHIELD))
		Set_Shield();		
	else if  ( (!FEATURE_RPT.OUT.Get_nSet) && (FEATURE_RPT.OUT.ParamSize == 1) &&
	      (FEATURE_RPT.OUT.Index_LSB == PAR8_FINGER_THRESHOLD))
		Set_Finger_Threshold();
	else if  ( (!FEATURE_RPT.OUT.Get_nSet) && (FEATURE_RPT.OUT.ParamSize == 1) &&
	      ((FEATURE_RPT.OUT.Index_LSB == PAR8_RESOLUTION)||(FEATURE_RPT.OUT.Index_LSB == PAR8_SCAN_SPEED)))
		Set_Resolution();
	else if  ( (!FEATURE_RPT.OUT.Get_nSet) && (FEATURE_RPT.OUT.ParamSize == 2) &&
	      ((FEATURE_RPT.OUT.Index_LSB >= PAR16_SET_BASELINE_0)&&(FEATURE_RPT.OUT.Index_LSB <= PAR16_SET_BASELINE_3)))
		Set_BaseLine();		
}

//-----------------------------------------------------------------------------
// void Sensor_FSM_Measure(void)
//
//	finite state machine to control the measurement process for 
//  the Sensor
//
//  IMPORTANT: 	call this routine once every millisecond. Otherwise the timing
//				parameters cannot be met.
//-----------------------------------------------------------------------------
void Sensor_FSM_Measure() 
{
	union uBigEndian result;
	BYTE BaseLineOn;
	BYTE ElapsedTime = SOF_Flag;
	SOF_Flag = 0;	//Has to be done directly after reading the Elapsed Time
	// decrement millisecond counter
	if (Countdown_1ms >= ElapsedTime)
		Countdown_1ms -= ElapsedTime;
	else if(Countdown_1ms >0)
		Countdown_1ms = 0;
	// process actual state machine
	switch (FSM)
	{
		case eIdle:
			if (Countdown_1ms == 0)
			{
				//LED1_Data_ADDR &= 0x07;
				Get_Param_16(PAR16_SAMPLERATE_MS, VOLATILE, &Countdown_1ms);
				CSD_1_ScanAllSensors();
				Get_Param_8(PAR8_AUTO_BASE_LINE, VOLATILE, &BaseLineOn);
				if(BaseLineOn)
				{
					CSD_1_UpdateAllBaselines();	//Updates Baseline and Calculate Difference Value
					CSD_1_bIsAnySensorActive();	//Updates the baSnsOnMask
				}
				else
				{
					DetectAllButtons();	//Updates the baSnsOnMask without calling previous UpdateAllBaselines()
				}
				result.w = CSD_1_waSnsResult[0];
				Report.CH2_MSB = result.BE.Hi;
				Report.CH2_LSB = result.BE.Lo;
				
				//result.w = CSD_1_waSnsDiff[2];
				result.w = CSD_1_waSnsResult[1];
				Report.CH3_MSB = result.BE.Hi;
				Report.CH3_LSB = result.BE.Lo;
				
				//result.w = CSD_1_waSnsBaseline[2];
				result.w = CSD_1_waSnsResult[2];
				Report.CH4_MSB = result.BE.Hi;
				Report.CH4_LSB = result.BE.Lo;
				
				//result.w = CSD_1_waSnsResult[2];
				result.w = CSD_1_waSnsResult[3];
				Report.CH5_MSB = result.BE.Hi;
				Report.CH5_LSB = result.BE.Lo;
				
				//LED1_Data_ADDR = ~CSD_1_baSnsOnMask[0];
				//LED1_Data_ADDR = (~CSD_1_baSnsOnMask[0])| 0x08;
				PRT5DR = CSD_1_baSnsOnMask[0]& 0x0F;
				Report.CH1 = CSD_1_baSnsOnMask[0];
				NewSample = TRUE;
			}
			break;			
	}
}

//-----------------------------------------------------------------------------
//	Toradex Algorithm for detecting the keys
//-----------------------------------------------------------------------------
//#define CSD_1_Debounce 3
//#define CSD_1_HYSTERESIS 10
void DetectAllButtons(void)
{
	BYTE i, Keys = 0;
	//WORD DiffVal;
	for(i=0;i<CSD_1_TotalSensorCount;i++)
	{
		//DiffVal = CSD_1_waSnsResult[i];
		if (CSD_1_waSnsResult[i] > CSD_1_waSnsBaseline[i] + CSD_1_NOISE_THRESHOLD)
		{
			CSD_1_waSnsDiff[i]= CSD_1_waSnsResult[i] - CSD_1_waSnsBaseline[i];
		}
		else
		{
			CSD_1_waSnsDiff[i] = 0;
		}
		switch(FSM_Keys[i])
		{
			case eOff:
				if(CSD_1_waSnsDiff[i] > CSD_1_baBtnFThreshold[i]+CSD_1_HYSTERESIS)
				{
					if(DebounceCount[i] > 1)
					{
						DebounceCount[i]--;
						FSM_Keys[i] = eGoOn;
					}
					else
					{
						DebounceCount[i] = CSD_1_DEBOUNCE;
						FSM_Keys[i] = eOn;
						Keys |= (0x01<<i);
					}
				}
			break;
			case eGoOn:
				if(CSD_1_waSnsDiff[i] > CSD_1_baBtnFThreshold[i]+CSD_1_HYSTERESIS)
				{
					if(DebounceCount[i] > 1)
					{
						DebounceCount[i]--;
					}
					else
					{
						DebounceCount[i] = CSD_1_DEBOUNCE;
						FSM_Keys[i] = eOn;
						Keys |= (0x01<<i);
					}
				}
				else if (CSD_1_waSnsDiff[i] < CSD_1_baBtnFThreshold[i]-CSD_1_HYSTERESIS)
				{
					DebounceCount[i] = CSD_1_DEBOUNCE;
					FSM_Keys[i] = eOff;
				}
			break;
			case eOn:
				Keys |= (0x01<<i);
				if(CSD_1_waSnsDiff[i] < CSD_1_baBtnFThreshold[i]-CSD_1_HYSTERESIS)
				{
					if(DebounceCount[i] > 1)
					{
						DebounceCount[i]--;
						FSM_Keys[i] = eGoOff;
					}
					else
					{
						DebounceCount[i] = CSD_1_DEBOUNCE;
						FSM_Keys[i] = eOff;
						Keys &= ~(0x01<<i);
					}
				}
			break;
			case eGoOff:
				Keys |= (0x01<<i);
				if(CSD_1_waSnsDiff[i] < CSD_1_baBtnFThreshold[i]-CSD_1_HYSTERESIS)
				{
					if(DebounceCount[i] > 1)
					{
						DebounceCount[i]--;
					}
					else
					{
						DebounceCount[i] = CSD_1_DEBOUNCE;
						FSM_Keys[i] = eOff;
						Keys &= ~(0x01<<i);
					}
				}
				else if (CSD_1_waSnsDiff[i] > CSD_1_baBtnFThreshold[i]+CSD_1_HYSTERESIS)
				{
					DebounceCount[i] = CSD_1_DEBOUNCE;
					FSM_Keys[i] = eOn;
				}
			break;
		}
	}
	CSD_1_baSnsOnMask[0]=Keys;
}

//-----------------------------------------------------------------------------
void Sensor_Debug(void)
{
/* currently no debug function */
}